/* main.c - df3tcl */

/* cf getopt_long_only(3) */
#define _GNU_SOURCE

#include "df3tcl.h"

#include <assert.h>
#include <getopt.h>
#include <locale.h>
#include <strings.h>


/* default working directory */
#define WORK_DIR "."  

/* "regular" */
#define APPPANIC appPanic(__FILE__,__LINE__)

/* who we are + CLI stuff */
char const * const app_name = "df3tcl";
char const * const app_version = "0.2.4";

static struct option app_opts_long[] = {
  {"diagnostic", no_argument,       0, 'D'},
  {"directory",  required_argument, 0, 'd'},
  {"file",       required_argument, 0, 'f'},
  {"help",       no_argument,       0, 'h'},
  {"safemode",   no_argument,       0, 's'},
  {"version",    no_argument,       0, 'V'},
  {"writemode",  no_argument,       0, 'w'},
  {NULL, 0, 0, CHRNUL}
};

static char const * const app_opts_short = "+Dd:f:hsVw";
// TODO TBIs.
static char const * const app_opts_help[] = {
  "\t display various \"debug\" messages",
  "path\t set output directory",
  "file[.tcl]\t script to run",
  "\t\t display this message, and exit",
  "\t run in safe interpreter.  TBI",
  "\t display the program version, and exit",
  "\t permit writing existing files",
  NULL
};

bool app_diag;
bool app_safemode;
char *app_script;

static char **data_files;

/* Tcl "package" interface.  for now, just one (regular) interpreter */
static Tcl_Interp *app_interp;

static Df3tclData app_state;


/* ------------------------------------------------------------------------- */

void appTclPanic(char const * const fn, int const ln) {
  int const e = errno;
  FREENULL(app_state.workdir);
  FREENULL(app_script);
  fprintf(stderr, "* ** *** aborted *** ** *\n"
                  "%s: Tcl \"panic\" in '%s' line %d.\n"
                  "libc reports e:%d \"%s\".\n",
          app_name, fn, ln, e, strerror(e));
  abort();
}

/* cf D Levine */
void appPanic(char const * const fn, int const ln) {
  int const e = errno;
  fprintf(stderr, "* ** *** aborted *** ** *\n"
                  "%s: \"cannot happen\" error in '%s' line %d,\n"
                  "libc reports e:%d \"%s\".\n",
          app_name, fn, ln, e, strerror(e));
  abort();
}

static void usageExit(int const) __attribute__((noreturn));
static void usageExit(int const ec) {
  char const * const fmt1 = "%s - DF3 Tcl Interpreter (%s).\n"
          "\n  usage: %s [option(s)] [file[.df3] ...]\n"
          "\noptions:\n";
  printf(fmt1, app_name, app_version, app_name);
  for (size_t i = 0; app_opts_long[i].name; ++i)
    printf("  -%c | -%s %s.\n",
            app_opts_long[i].val, app_opts_long[i].name, app_opts_help[i]);
  char const * const fmt2 = "\nsee 'man 1 %s' for details.\n";
  printf(fmt2, app_name);
  exit(ec);
}

static void versionExit(void) __attribute__((noreturn));
static void versionExit(void) {
  char const * const fmt = "%s %s\ncopyright (c) 2020 jr.\n"
          "license GPLv3+ <http://gnu.org/licenses/gpl.html>\n"
          "this is free software, you are free to .. whatever.\n"
          "there is no warranty.\n";
  printf(fmt, app_name, app_version);
  exit(EXIT_SUCCESS);
}


/* ------------------------------------------------------------------------- */

/* the 'flag' arg tells whether extension the problem, or out-of-memory */
char *storeFname(char const * const fname,
                 char const * const extn,
                 bool *flag) {
  char buf[PATH_MAX] = {CHRNUL};

  assert((PATH_MAX - 4) > strlen(fname));
  strcpy(buf, fname);

  char *p = strrchr(buf, CHRDOT);
  if (!p  ||  4 != strlen(p))
    strcat(buf, extn);
  else if (strcmp(p, extn)) {
    *flag = false;
    return NULL;
  }

  return strdup(buf);
}


/* ------------------------------------------------------------------------- */

static bool argDirectory(char const * const arg) {
  size_t n = strlen(arg);
  /* POSIX guarantees file name length 14 chars, plus separator */
  if (!n  ||  (PATH_MAX - 15) < n)
    return false;
  else if (CHRPTS == arg[n - 1])
    n -= 1;

  if (!(app_state.workdir = strndup(arg, n)))
    APPPANIC;

  return true;
}

static bool argFile(char const * const arg) {
  bool f = true;
  app_script = storeFname(arg, ".tcl", &f);
  if (!f) {
    fprintf(stderr, "%s: oops, can't use file name '%s'.\n", app_name, arg);
    return false;
  }
  else if (!app_script)
    APPPANIC;
  return true;
}

static void processCmdLine(int ac, char *av[]) {
  bool havediag, havesafe, havescript, havewrite, havewdir;

  havediag = havesafe = havescript = havewrite = havewdir = false;

  opterr = 0;
  while (1) {
    int idx = 0;
    int c = getopt_long_only(ac, av, app_opts_short, app_opts_long, &idx);
    if (-1 == c) break;
    switch (c) {
      case 'D':
        if (havediag) usageExit(2);
        havediag = app_diag = true;
        break;
      case 'd':
        if (havewdir) usageExit(2);
        else if (!argDirectory(optarg)) exit(EXIT_FAILURE);
        havewdir = true;
        break;
      case 'f':
        if (havescript) usageExit(2);
        else if (!argFile(optarg)) exit(EXIT_FAILURE);
        havescript = true;
        break;
      case 'h': usageExit(0);
      case 's':
        if (havesafe) usageExit(2);
        havesafe = app_safemode = true;
        break;
      case 'V': versionExit();
      case 'w':
        if (havewrite) usageExit(2);
        havewrite = app_state.writemode = true;
        break;
      default: usageExit(2);
    }
  }

  if (!havescript) {
    fprintf(stderr, "%s: oops, require option '-file script[.tcl]'.\n",
            app_name);
    exit(EXIT_FAILURE);
  }

  if (!havewdir)
    if (!(app_state.workdir = strdup(WORK_DIR)))
      APPPANIC;

  /* remaining args */
  if (ac - optind)
    data_files = &av[optind];

  return;
}


/* ------------------------------------------------------------------------- */

static void tidyup(void) {
  Tcl_Finalize();
  FREENULL(app_state.workdir);
  FREENULL(app_script);
  return;
}

static void setup(char const * const self) {
  if (!setlocale(LC_ALL, "POSIX"))       APPPANIC;
  else if (setreuid(getuid(), getuid())) APPPANIC;
  else if (setregid(getgid(), getgid())) APPPANIC;
  else if (atexit(tidyup))               APPPANIC;
  Tcl_FindExecutable(self);
  app_state.nextid = 1;
  app_state.writemode = false;
  app_safemode = false;
  return;
}


int main(int argc, char *argv[]) {
  setup(argv[0]);
  processCmdLine(argc, argv);

  if (!(app_interp = Tcl_CreateInterp())) {
    fprintf(stderr, "%s: oops, Tcl interpreter creation failed.\n", app_name);
    exit(EXIT_FAILURE);
  }

// TODO "safe" slave.  cf std(err|in|out).

  prepTclif(&app_state, app_interp, data_files);

  int const rtn = Tcl_EvalFile(app_interp, app_script);
  char const * const res = Tcl_GetStringResult(app_interp);
  if (TCL_OK != rtn) {
    fprintf(stderr, "%s: interp error: '%s'.\n", app_name, res);
    Tcl_Exit(EXIT_FAILURE);
  }

  if (app_diag)
    fprintf(stderr, "[II] interp result: '%s'.\n", res);

  Tcl_DeleteInterp(app_interp);

  return EXIT_SUCCESS;
}

